import logging
import os
from urllib.parse import urlsplit, urlunsplit

from ._utils import url_join

LOGGER = logging.getLogger(__name__)


def get_workspace() -> str:
    try:
        return os.environ["RC_WORKSPACE_ID"]
    except KeyError as err:
        raise RuntimeError("Missing environment variable 'RC_WORKSPACE_ID'") from err


def get_endpoint() -> str:
    if endpoint := os.getenv("RC_API_URL_V1"):
        return endpoint

    # Note (2023-06-28):
    # While Control Room does set RC_API_URL_V1, this is not currently supported
    # in local development. We can use as fallback the Vault endpoint, but we
    # need to replace some parts to get the correct value. Examples:
    #   api.eu1.robocloud.eu -> api.eu1.robocorp.com/v1
    #   api.ci.robocloud.dev -> api.ci.robocorp.dev/v1
    #   api.eu2.robocorp.com -> api.eu2.robocorp.com/v1
    #   api.us1.robocorp.com -> api.us1.robocorp.com/v1
    LOGGER.info("Missing environment variable 'RC_API_URL_V1', attempting fallback")
    endpoint = os.getenv("RC_API_SECRET_HOST")
    if endpoint is None:
        raise RuntimeError("Missing environment variable 'RC_API_URL_V1'")

    url = urlsplit(endpoint)

    # Replace hostname to match correct API
    netloc = url.netloc
    if url.hostname and url.hostname.endswith("robocloud.eu"):
        netloc = netloc.replace("robocloud.eu", "robocorp.com")
    elif url.hostname and url.hostname.endswith("robocloud.dev"):
        netloc = netloc.replace("robocloud.dev", "robocorp.dev")
    url = url._replace(netloc=netloc)

    # Append /v1/ to path if not already there
    if not url.path.rstrip("/").endswith("v1"):
        path = url_join(url.path, "v1")
        url = url._replace(path=path)

    result = urlunsplit(url)
    LOGGER.info("Adapted endpoint: %s -> %s", endpoint, result)
    return result


def get_token() -> str:
    if token := os.getenv("RC_API_TOKEN_V1"):
        return token

    # Note (2023-06-28):
    # While Control Room does set RC_API_TOKEN_V1, this is not currently
    # supported in local development. We can use as fallback the Vault token
    # which is identical.
    LOGGER.info("Missing environment variable 'RC_API_TOKEN_V1', attempting fallback")
    if token := os.getenv("RC_API_SECRET_TOKEN"):
        return token

    raise RuntimeError("Missing environment variable 'RC_API_TOKEN_V1'")



import base64
import logging
import math
import time
from typing import Optional, cast
from urllib.parse import quote_plus

from ._requests import HTTPError, Requests, Response
from ._types import (
    Asset,
    AssetDetails,
    AssetUploadResponse,
    AssetUploadState,
    ContentType,
)
from ._utils import url_join

LOGGER = logging.getLogger(__name__)

# Assets under `DATA_LIMIT` can be uploaded directly through the POST request,
# instead of requesting a separate upload url
DATA_LIMIT = 5 * 10**6  # 5MB


def _estimate_base64_size(size: int) -> int:
    """Return guess for content length after being base64 encoded."""
    return math.ceil(size / 3) * 4


class AssetNotFound(HTTPError):
    """No asset with given name/id found."""


class AssetUploadFailed(RuntimeError):
    """There was an unexpected error while uploading an asset."""


class AssetsClient:
    def __init__(self, workspace: str, endpoint: str, token: str):
        self._endpoint = endpoint
        self._workspace = workspace
        self._token = token

        route_prefix = url_join(self._endpoint, "workspaces", self._workspace, "assets")
        default_headers = {
            "Authorization": f"Bearer {self._token}",
            "Content-Type": "application/json",
        }

        self._client = Requests(route_prefix, default_headers=default_headers)

    def _handle_asset_not_found(self, asset_id: str):
        """Create an error handler that raises `AssetNotFound` for 404 status codes."""

        def _handle_error(response: Response):
            try:
                self._client.handle_error(response)
            except HTTPError as exc:
                if exc.status_code == 404:
                    # NOTE(cmin764): Any `404 Not Found` error won't be retried
                    #  and will bubble-up from the first request.
                    raise AssetNotFound(
                        f"Asset with id {asset_id!r} not found",
                        status_code=exc.status_code,
                        reason=exc.reason,
                    ) from exc
                raise

        return _handle_error

    def list_assets(self) -> list[Asset]:
        """Retrieve list of assets."""
        return self._client.get("").json()

    def create_asset(self, name: str) -> AssetDetails:
        """Create a new asset."""
        payload = {"name": name}
        response = self._client.post("", json=payload)
        return cast(AssetDetails, response.json())

    def get_asset(self, asset_id: str) -> AssetDetails:
        """Returns the details of a single asset, including its payload."""
        path = quote_plus(asset_id)
        handler = self._handle_asset_not_found(asset_id)
        response = self._client.get(path, _handle_error=handler)
        return cast(AssetDetails, response.json())

    def delete_asset(self, asset_id: str):
        """Delete the given asset."""
        path = quote_plus(asset_id)
        handler = self._handle_asset_not_found(asset_id)
        self._client.delete(path, _handle_error=handler)

    def upload_asset(
        self,
        asset_id: str,
        content: bytes,
        content_type: Optional[ContentType] = None,
        wait: bool = True,
        poll_interval: float = 0.5,
    ):
        """Upload an asset payload, and optionally wait for it to finish."""
        if _estimate_base64_size(len(content)) < DATA_LIMIT:
            LOGGER.info("Content size under data limit, uploading directly")
            data_content = base64.b64encode(content).decode("ascii")
            response = self._create_upload(
                asset_id,
                content_type=content_type,
                data=data_content,
            )
        else:
            LOGGER.info("Content size over data limit, requesting upload URL")
            response = self._create_upload(asset_id, content_type=content_type)
            self._client.put(response["upload_url"], data=content, headers={})

        # NOTE (2023-07-04):
        # The asset upload should be done if the response says so,
        # but in reality it sometimes takes a while for the value to propagate

        # if response["status"] == "completed":
        #    return

        if not wait:
            return

        LOGGER.info("Waiting for asset upload to complete")
        while True:
            state = self._get_upload(asset_id, response["id"])
            status = state["status"]

            if status == "pending":
                time.sleep(poll_interval)
                continue
            elif status == "completed":
                break
            elif status == "failed":
                reason = state.get("reason", "unknown")
                raise AssetUploadFailed(f"Asset upload failed: {reason}")
            else:
                raise AssetUploadFailed(f"Asset upload invalid status: {status}")

    def _create_upload(
        self,
        asset_id: str,
        content_type: Optional[ContentType] = None,
        data: Optional[str] = None,
    ) -> AssetUploadResponse:
        """Create an upload for an existing asset.

        For payloads of less than 5MB you can upload the contents directly on
        upload creation. For larger payloads you can create the upload and use
        the returned upload URL for delivering the content.
        """
        payload = {}
        if content_type is not None:
            payload["content_type"] = content_type
        if data is not None:
            payload["data"] = data

        path = url_join(quote_plus(asset_id), "upload")
        handler = self._handle_asset_not_found(asset_id)
        response = self._client.post(path, json=payload, _handle_error=handler)
        return cast(AssetUploadResponse, response.json())

    def _get_upload(self, asset_id: str, upload_id: str) -> AssetUploadState:
        """Return the details of a single asset upload."""
        path = url_join(quote_plus(asset_id), "uploads", upload_id)
        response = self._client.get(path)
        return cast(AssetUploadState, response.json())



import json
import logging
import os
import random
import time
import urllib.parse
from typing import Callable, Optional

import requests
from requests.exceptions import HTTPError as _HTTPError
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception,
    stop_after_attempt,
    wait_random_exponential,
)

LOGGER = logging.getLogger(__name__)
DEBUG = bool(os.getenv("RPA_DEBUG_API"))


def _needs_retry(exc: BaseException) -> bool:
    # Don't retry on some specific error codes or messages.

    # https://www.restapitutorial.com/httpstatuscodes.html
    # 400 - payload is bad and needs to be changed
    # 401 - missing auth bearer token
    # 403 - auth is in place, but not allowed (insufficient privileges)
    # 404 - resource does not exist
    # 409 - payload not good for the affected resource
    no_retry_codes = [400, 401, 403, 404, 409]

    if isinstance(exc, HTTPError):
        if exc.status_code in no_retry_codes:
            return False

        if exc.status_code == 429:
            # We hit the rate limiter, so sleep extra before retrying.
            seconds = round(random.uniform(1, 3), 2)
            LOGGER.warning("Rate limit hit, sleeping %.2fs...", seconds)
            time.sleep(seconds)

    return True


def _before_sleep_log():
    log_method = LOGGER.log

    def extensive_log(level, msg, *args, **kwargs):
        log_method(level, msg, *args, **kwargs)
        if DEBUG:
            LOGGER.debug(msg, *args)

    # Monkeypatch inner logging function so it produces an exhaustive log when
    # used under the before-sleep logging utility in `tenacity`.
    LOGGER.log = extensive_log
    return before_sleep_log(LOGGER, logging.DEBUG, exc_info=True)


# Exported for ease-of-use
Response = requests.Response


class HTTPError(_HTTPError):
    """Custom `requests` HTTP error with status code and reason"""

    def __init__(self, message: str, status_code: int, reason: str):
        super().__init__(message)
        self.message = message
        self.status_code = status_code
        self.reason = reason

    def __str__(self):
        return f"{self.status_code} {self.reason} - {self.message}"


class Requests:
    """Wrapper over `requests` 3rd-party with error handling and retrying support."""

    def __init__(self, route_prefix: str = "", default_headers: Optional[dict] = None):
        self._route_prefix = route_prefix
        self._default_headers = default_headers

    def handle_error(self, response: requests.Response):
        http_status = f"{response.status_code} {response.reason!r}"
        if 500 <= response.status_code < 600:
            LOGGER.critical("Server error: %s", http_status)
        elif 400 <= response.status_code < 500:
            LOGGER.info("Client error: %s", http_status)
        else:
            LOGGER.debug("Response: %s", http_status)

        if response.ok:
            return

        # Fallback values
        status_code = response.status_code
        reason = response.reason
        message = response.text

        try:
            fields = response.json()

            # For some reason we might still get a string from the deserialized
            # JSON payload, possible due to some double encoding bug in CR
            while not isinstance(fields, dict):
                fields = json.loads(fields)

            if "status" in fields:
                status_code = int(fields["status"])

            if "code" in fields:
                reason = fields["code"]
            elif "error" in fields and "code" in fields["error"]:
                reason = fields["error"]["code"]

            if "message" in fields:
                message = fields["message"]
            elif "error" in fields and "message" in fields["error"]:
                message = fields["error"]["message"]
        except Exception as exc:
            LOGGER.critical("Failed to parse error response: %r", exc)

        raise HTTPError(message, status_code=status_code, reason=reason)

    @retry(
        # Retry until either succeed or trying for the fifth time and still failing.
        # So sleep and retry for 4 times at most.
        stop=stop_after_attempt(5),
        # If the exception is no worth retrying or the number of tries is depleted,
        # then re-raise the last raised exception.
        reraise=True,
        # Decide if the raised exception needs retrying or not.
        retry=retry_if_exception(_needs_retry),
        # Produce debugging logging prior to each time we sleep & re-try.
        before_sleep=_before_sleep_log(),
        # Sleep between the tries with a random float amount of seconds like so:
        # 1. [0, 2]
        # 2. [0, 4]
        # 3. [0, 5]
        # 4. [0, 5]
        wait=wait_random_exponential(multiplier=2, max=5),
    )
    def _request(
        self,
        verb: Callable[..., requests.Response],
        url: str,
        *args,
        _handle_error: Optional[Callable[[requests.Response], None]] = None,
        _sensitive: bool = False,
        headers: Optional[dict] = None,
        **kwargs,
    ) -> requests.Response:
        # Absolute URLs override the prefix, so they are safe to be sent
        # as they'll be the same after joining.
        url = urllib.parse.urljoin(self._route_prefix, url)
        headers = headers if headers is not None else self._default_headers
        handle_error = _handle_error or self.handle_error

        log_url = url
        if _sensitive:
            # Omit query from the URL since might contain sensitive info.
            split = urllib.parse.urlsplit(log_url)
            split = split._replace(query="")
            log_url = urllib.parse.urlunsplit(split)

        LOGGER.debug("%s %r", verb.__name__.upper(), log_url)
        response = verb(url, *args, headers=headers, **kwargs)
        handle_error(response)
        return response

    def post(self, *args, **kwargs) -> requests.Response:
        return self._request(requests.post, *args, **kwargs)

    def get(self, *args, **kwargs) -> requests.Response:
        return self._request(requests.get, *args, **kwargs)

    def put(self, *args, **kwargs) -> requests.Response:
        return self._request(requests.put, *args, **kwargs)

    def delete(self, *args, **kwargs) -> requests.Response:
        return self._request(requests.delete, *args, **kwargs)



import sys
from typing import Literal, Optional, Union

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


ContentType = str

Asset = TypedDict(
    "Asset",
    {
        "id": str,
        "name": str,
    },
)

AssetUploadRequest = TypedDict(
    "AssetUploadRequest",
    {
        "content_type": ContentType,
        "data": Optional[str],
    },
)

AssetUploadResponse = TypedDict(
    "AssetUploadResponse",
    {
        "id": str,
        "status": Union[Literal["pending"], Literal["completed"]],
        "content_type": ContentType,
        "upload_url": Optional[str],
    },
)

AssetUploadPending = TypedDict(
    "AssetUploadPending",
    {
        "id": str,
        "status": Literal["pending"],
        "content_type": ContentType,
    },
)

AssetUploadCompleted = TypedDict(
    "AssetUploadCompleted",
    {
        "id": str,
        "status": Literal["completed"],
        "content_type": ContentType,
    },
)

AssetUploadFailed = TypedDict(
    "AssetUploadFailed",
    {
        "id": str,
        "status": Literal["failed"],
        "content_type": ContentType,
        "reason": Optional[str],
    },
)

AssetUploadState = Union[
    AssetUploadPending,
    AssetUploadCompleted,
    AssetUploadFailed,
]

AssetPayloadEmpty = TypedDict(
    "AssetPayloadEmpty",
    {
        "type": Literal["empty"],
    },
)

AssetPayloadData = TypedDict(
    "AssetPayloadData",
    {
        "type": Literal["data"],
        "content_type": ContentType,
        "data": str,
    },
)

AssetPayloadUrl = TypedDict(
    "AssetPayloadUrl",
    {
        "type": Literal["url"],
        "content_type": ContentType,
        "url": str,
    },
)

AssetDetails = TypedDict(
    "AssetDetails",
    {
        "id": str,
        "name": str,
        "payload": Union[AssetPayloadEmpty, AssetPayloadData, AssetPayloadUrl],
    },
)



import json
import logging
import mimetypes
import os
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

from ._client import AssetNotFound, AssetUploadFailed

if TYPE_CHECKING:
    from ._client import AssetsClient
    from ._requests import Response

__version__ = "0.3.1"
version_info = [int(x) for x in __version__.split(".")]

JSON = Union[dict[str, "JSON"], list["JSON"], str, int, float, bool, None]

LOGGER = logging.getLogger(__name__)

# Known (additional) mimetypes from file extensions
KNOWN_MIMETYPES = [
    ("text/x-yaml", ".yml"),
    ("text/x-yaml", ".yaml"),
]


@lru_cache(maxsize=1)
def _get_client() -> "AssetsClient":
    """
    Creates and returns an Asset Storage API client based on the injected
    environment variables from Control Room (or RCC).
    """
    from ._client import AssetsClient
    from ._environment import get_endpoint, get_token, get_workspace

    workspace = get_workspace()
    endpoint = get_endpoint()
    token = get_token()

    return AssetsClient(workspace, endpoint, token)


def list_assets() -> list[str]:
    """List all the existing assets.

    Returns:
        A list of available assets' names
    """
    return [asset["name"] for asset in _get_client().list_assets()]


def delete_asset(name: str):
    """Delete an asset by providing its `name`.

    This operation cannot be undone.

    Args:
        name: Asset to delete

    Raises:
        AssetNotFound: Asset with the given name does not exist
    """
    LOGGER.info("Deleting asset: %s", name)
    client = _get_client()
    client.delete_asset(asset_id=f"name:{name}")


def _get_asset(name: str) -> "Response":
    """Get an asset's payload URL."""
    from ._requests import Requests

    client = _get_client()
    details = client.get_asset(asset_id=f"name:{name}")
    payload = details["payload"]

    if payload["type"] == "url":
        url = payload["url"]
        return Requests().get(url)
    if payload["type"] == "empty":
        raise ValueError(
            f"Asset {details['name']!r} is empty."
            + " It could mean an upload is still pending"
            + " or has previously failed."
        )
    else:
        # Note (2023-07-04):
        # The 'data' payload type should only be used when uploading,
        # and it should never be in the response when getting an asset.
        raise RuntimeError(f"Unsupported payload type: {payload['type']}")


def get_text(name: str) -> str:
    """Return the given asset as text.

    Arguments:
        name: Name of asset

    Returns:
        Asset content as text

    Raises:
        AssetNotFound: No asset defined with given name
    """
    response = _get_asset(name)
    return response.text


def get_json(name: str, **kwargs) -> JSON:
    """Return the given asset as a deserialized JSON object.

    Arguments:
        name: Name of asset
        **kwargs: Additional parameters for `json.loads`

    Returns:
        Asset content as a Python object (dict, list etc.)

    Raises:
        AssetNotFound: No asset defined with given name
        JSONDecodeError: Asset was not valid JSON
    """
    response = _get_asset(name)
    return response.json(**kwargs)


def get_file(name: str, path: Union[os.PathLike, str], exist_ok=False) -> Path:
    """Fetch the given asset and store it in a file.

    Arguments:
        name: Name of asset
        path: Destination path for downloaded file
        exist_ok: Overwrite file if it already exists

    Returns:
        Path to created file

    Raises:
        AssetNotFound: No asset defined with given name
        FileExistsError: Destination already exists
    """
    response = _get_asset(name)

    path = Path(path).absolute()
    if path.exists() and not exist_ok:
        raise FileExistsError(f"File already exists: {path}")

    path.write_bytes(response.content)
    return path


def get_bytes(name: str) -> bytes:
    """Return the given asset as bytes.

    Arguments:
        name: Name of asset

    Returns:
        Asset content as bytes

    Raises:
        AssetNotFound: No asset defined with given name
    """
    response = _get_asset(name)
    return response.content


def _set_asset(name: str, content: bytes, content_type: str, wait: bool):
    """Upload asset content, and create asset if it doesn't already exist."""
    client = _get_client()

    try:
        details = client.get_asset(asset_id=f"name:{name}")
        LOGGER.debug("Updating existing asset with id: %s", details["id"])
    except AssetNotFound:
        details = client.create_asset(name=name)
        LOGGER.debug("Created new asset with id: %s", details["id"])

    LOGGER.info("Uploading asset %r (content-type: %s)", name, content_type)
    client.upload_asset(details["id"], content, content_type, wait)


def set_text(name: str, text: str, wait: bool = True):
    """Create or update an asset to contain the given string.

    Arguments:
        name: Name of asset
        text: Text content for asset
        wait: Wait for asset to update
    """
    content = text.encode("utf-8")
    content_type = "text/plain"
    _set_asset(name, content, content_type, wait)


def set_json(name: str, value: JSON, wait: bool = True, **kwargs):
    """Create or update an asset to contain the given object, serialized as JSON.

    Arguments:
        name: Name of asset
        value: Value for asset, e.g. dict or list
        wait: Wait for asset to update
        **kwargs: Additional arguments for `json.dumps`
    """
    content = json.dumps(value, **kwargs).encode("utf-8")
    content_type = "application/json"
    _set_asset(name, content, content_type, wait)


def set_file(
    name: str,
    path: Union[os.PathLike, str],
    content_type: Optional[str] = None,
    wait: bool = True,
):
    """Create or update an asset to contain the contents of the given file.

    Arguments:
        name: Name of asset
        path: Path to file
        content_type: Content type (or mimetype) of file, detected automatically
          from file extension if not defined
        wait: Wait for asset to update
    """
    if content_type is None:
        for type_, ext in KNOWN_MIMETYPES:
            mimetypes.add_type(type_, ext)

        content_type, _ = mimetypes.guess_type(path)
        if content_type is not None:
            LOGGER.info("Detected content type %r", content_type)
        else:
            content_type = "application/octet-stream"
            LOGGER.info("Unable to detect content type, using %r", content_type)

    content = Path(path).read_bytes()
    _set_asset(name, content, content_type, wait)


def set_bytes(
    name: str,
    data: bytes,
    content_type="application/octet-stream",
    wait: bool = True,
):
    """Create or update an asset to contain the given bytes.

    Arguments:
        name: Name of asset
        data: Raw content
        content_type: Content type (or mimetype) of asset
        wait: Wait for asset to update
    """
    _set_asset(name, data, content_type, wait)


__all__ = [
    "AssetNotFound",
    "AssetUploadFailed",
    "list_assets",
    "delete_asset",
    "get_text",
    "get_json",
    "get_file",
    "get_bytes",
    "set_text",
    "set_json",
    "set_file",
    "set_bytes",
]
